
# Copyright 2016-2018 Euratom
# Copyright 2016-2018 United Kingdom Atomic Energy Authority
# Copyright 2016-2018 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.

import os
import numpy as np
from scipy.constants import elementary_charge

from cherab.core.utility import PhotonToJ
from cherab.core.atomic.elements import lookup_isotope, hydrogen, deuterium, tritium
from cherab.openadas import OpenADAS

from cherab.solps.eirene import load_fort44_file
from cherab.solps.b2.parse_b2_block_file import load_b2f_file
from cherab.solps.mesh_geometry import SOLPSMesh
from cherab.solps.solps_plasma import SOLPSSimulation, prefer_element, eirene_flux_to_velocity, b2_flux_to_velocity


# Code based on script by Felix Reimold (2016)
def load_solps_from_raw_output(simulation_path='', debug=False, mesh_file_path=None,
                               b2_state_file_path=None, b2_plasma_file_path=None,
                               eirene_fort44_file_path=None, atomic_data=None):
    """
    Load a SOLPS simulation from raw SOLPS output files.

    Relevant files are:
    * mesh description file (b2fgmtry)
    * B2 plasma state (b2fstate)
    * B2 plasma solution, formatted (b2fplasmf), optional
    * Eirene output file (fort.44), optional

    :param str simulation_path: String path to simulation directory.
                                The SOLPS output files will be searched for in this directory
                                by their default names, unless the paths to specific files are
                                explicitly provided by the user.
                                Example: '/home/user/solps5/runs/simulation_name/run'.
                                Defaults to the current working directory.
    :param bool debug: A flag for displaying textual debugging information when parsing
                       the SOLPS files. Defaults to False.
    :param str mesh_file_path: String path to mesh description file (b2fgmtry).
                               Defaults to '{simulation_path}/b2fgmtry' if None.
    :param str b2_state_file_path: String path to B2 plasma state file (b2fstate).
                                   Defaults to '{simulation_path}/b2fstate' if None.
    :param str b2_plasma_file_path: String path to formatted B2 plasma solution file (b2fplasmf).
                                    Defaults to '{simulation_path}/b2fplasmf' if None.
    :param str eirene_fort44_file_path: String path to Eirene output file (fort.44).
                                        Defaults to '{simulation_path}/fort.44' if None.
    :param AtomicData atomic_data: The atomic data provider. Used here to convert
                                   the radiation density from photons/s/m3 to W/m3.
                                   Defaults to `OpenADAS()` if None.

    :rtype: SOLPSSimulation
    """

    mesh_file_path = mesh_file_path or os.path.join(simulation_path, 'b2fgmtry')
    b2_state_file_path = b2_state_file_path or os.path.join(simulation_path, 'b2fstate')
    b2_plasma_file_path = b2_plasma_file_path or os.path.join(simulation_path, 'b2fplasmf')
    eirene_fort44_file_path = eirene_fort44_file_path or os.path.join(simulation_path, 'fort.44')

    if not os.path.isfile(mesh_file_path):
        raise RuntimeError("No B2 mesh description file ({}) found.".format(mesh_file_path))
    _, _, geom_data_dict = load_b2f_file(mesh_file_path, debug=debug)

    if not os.path.isfile(b2_state_file_path):
        raise RuntimeError("No B2 plasma state file ({}) found.".format(b2_state_file_path))
    header_dict, sim_info_dict, mesh_data_dict = load_b2f_file(b2_state_file_path, debug=debug)

    if not os.path.isfile(b2_plasma_file_path):
        print("Warning! No B2 plasma solution formatted file ({}) found.".format(b2_plasma_file_path),
              "No total_radiation data will be available.")
        have_b2plasmf = False
    else:
        _, _, plasma_solution_dict = load_b2f_file(b2_plasma_file_path, debug=debug, header_dict=header_dict)
        have_b2plasmf = True

    if not os.path.isfile(eirene_fort44_file_path):
        print("Warning! No EIRENE output file ({}) found.".format(eirene_fort44_file_path),
              "Assuming B2 stand-alone simulation.")
        b2_standalone = True
    else:
        # Load data for neutral species from EIRENE output file
        eirene = load_fort44_file(eirene_fort44_file_path, debug=debug)
        b2_standalone = False

    mesh = create_mesh_from_geom_data(geom_data_dict)

    ny = mesh.ny  # radial
    nx = mesh.nx  # poloidal

    # Load each plasma species in simulation
    species_list = []
    neutral_indx = []
    hydrogen_neutrals = {}
    for i in range(len(sim_info_dict['zn'])):

        zn = int(sim_info_dict['zn'][i])  # Nuclear charge
        am = int(round(float(sim_info_dict['am'][i])))  # Atomic mass number
        charge = int(sim_info_dict['zamax'][i])  # Ionisation/charge
        isotope = lookup_isotope(zn, number=am)
        species = prefer_element(isotope)  # Prefer Element over Isotope if the mass number is the same
        species_list.append((species.name, charge))
        if charge == 0:  # updating neutral index
            neutral_indx.append(i)
            if species in (hydrogen, deuterium, tritium):
                hydrogen_neutrals[species] = i

    sim = SOLPSSimulation(mesh, species_list)

    # Load magnetic field
    sim.b_field = geom_data_dict['bb'][:3]
    # sim.b_field_cylindrical is created automatically

    # Load electron species
    sim.electron_temperature = mesh_data_dict['te'] / elementary_charge
    sim.electron_density = mesh_data_dict['ne']

    # Load ion temperature
    sim.ion_temperature = mesh_data_dict['ti'] / elementary_charge

    # Load species density
    sim.species_density = mesh_data_dict['na']

    # Load parallel velocity
    parallel_velocity = mesh_data_dict['ua']

    # Load poloidal and radial particle fluxes for velocity calculation
    poloidal_flux = mesh_data_dict['fna'][::2]
    radial_flux = mesh_data_dict['fna'][1::2]

    # Obtaining velocity from B2 flux
    sim.velocities_cylindrical = b2_flux_to_velocity(sim, poloidal_flux, radial_flux, parallel_velocity)

    if not b2_standalone:
        # Obtaining additional data from EIRENE and replacing data for neutrals
        # Note EIRENE data grid is slightly smaller than SOLPS grid, for example (98, 38) => (96, 36)
        # Need to pad EIRENE data to fit inside larger B2 array

        neutral_density = np.zeros((len(neutral_indx), ny, nx))
        neutral_density[:, 1:-1, 1:-1] = eirene.da
        sim.species_density[neutral_indx] = neutral_density

        # Obtaining neutral atom velocity from EIRENE flux
        # Note that if the output for fluxes was turned off, eirene.ppa and eirene.rpa are all zeros
        if np.any(eirene.ppa) or np.any(eirene.rpa):
            neutral_poloidal_flux = np.zeros((len(neutral_indx), ny, nx))
            neutral_poloidal_flux[:, 1:-1, 1:-1] = eirene.ppa

            neutral_radial_flux = np.zeros((len(neutral_indx), ny, nx))
            neutral_radial_flux[:, 1:-1, 1:-1] = eirene.rpa

            neutral_parallel_velocity = np.zeros((len(neutral_indx), ny, nx))  # must be zero outside EIRENE grid
            neutral_parallel_velocity[:, 1:-1, 1:-1] = parallel_velocity[neutral_indx, 1:-1, 1:-1]

            sim.velocities_cylindrical[neutral_indx] = eirene_flux_to_velocity(sim, neutral_poloidal_flux, neutral_radial_flux,
                                                                               neutral_parallel_velocity)
            sim.velocities_cylindrical = sim.velocities_cylindrical  # Updating sim.velocities

        # Obtaining neutral temperatures
        ta = np.zeros((eirene.ta.shape[0], ny, nx))
        ta[:, 1:-1, 1:-1] = eirene.ta
        # extrapolating
        for i in (0, -1):
            ta[:, i, 1:-1] = eirene.ta[:, i, :]
            ta[:, 1:-1, i] = eirene.ta[:, :, i]
        for i, j in ((0, 0), (0, -1), (-1, 0), (-1, -1)):
            ta[:, i, j] = eirene.ta[:, i, j]
        sim.neutral_temperature = ta / elementary_charge

        # Obtaining total radiation
        if have_b2plasmf and eirene.eradt is not None:
            line_radiation = plasma_solution_dict['rqrad'].sum(0)
            bremsstrahlung = plasma_solution_dict['rqbrm'].sum(0)
            eradt_raw_data = eirene.eradt.sum(0)
            total_radiation = line_radiation + bremsstrahlung
            total_radiation[1:-1, 1:-1] -= eradt_raw_data
            sim.total_radiation = total_radiation / mesh.vol

        # Obtaining molecular and total H-alpha radiation
        if len(hydrogen_neutrals) and (eirene.emism is not None or eirene.emist is not None):
            atomic_data = atomic_data or OpenADAS()
            total_hydrogen_density = sim.species_density[list(hydrogen_neutrals.values())].sum(0)
            effective_energy = 0
            for isotope, i in hydrogen_neutrals.items():
                wavelength = atomic_data.wavelength(isotope, 0, (3, 2))
                fraction = np.divide(sim.species_density[i], total_hydrogen_density,
                                     out=np.zeros_like(total_hydrogen_density),
                                     where=(total_hydrogen_density > 0))
                effective_energy += PhotonToJ.to(fraction, wavelength)

            if eirene.emism is not None:
                halpha_mol_radiation = np.zeros((ny, nx))
                halpha_mol_radiation[1:-1, 1:-1] = eirene.emism[0]
                sim.halpha_mol_radiation = effective_energy * halpha_mol_radiation

            if eirene.emist is not None:
                halpha_total_radiation = np.zeros((ny, nx))
                halpha_total_radiation[1:-1, 1:-1] = eirene.emist[0]
                sim.halpha_total_radiation = effective_energy * halpha_total_radiation

        sim.eirene_simulation = eirene

    return sim


def create_mesh_from_geom_data(geom_data):

    r = geom_data['crx']
    z = geom_data['cry']
    vol = geom_data['vol']

    # Loading neighbouring cell indices
    neighbix = np.zeros(r.shape, dtype=int)
    neighbiy = np.zeros(r.shape, dtype=int)

    neighbix[0] = geom_data['leftix'].astype(int)  # poloidal prev.
    neighbix[1] = geom_data['bottomix'].astype(int)  # radial prev.
    neighbix[2] = geom_data['rightix'].astype(int)  # poloidal next
    neighbix[3] = geom_data['topix'].astype(int)  # radial next

    neighbiy[0] = geom_data['leftiy'].astype(int)
    neighbiy[1] = geom_data['bottomiy'].astype(int)
    neighbiy[2] = geom_data['rightiy'].astype(int)
    neighbiy[3] = geom_data['topiy'].astype(int)

    # In SOLPS cell indexing starts with -1 (guarding cell), but in SOLPSMesh -1 means no neighbour.
    neighbix += 1
    neighbiy += 1
    neighbix[neighbix == r.shape[2]] = -1
    neighbiy[neighbiy == r.shape[1]] = -1

    mesh = SOLPSMesh(r, z, vol, neighbix, neighbiy)

    return mesh



# Copyright 2016-2018 Euratom
# Copyright 2016-2018 United Kingdom Atomic Energy Authority
# Copyright 2016-2018 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.


# Code based on script by Felix Reimold (2016)
class Eirene:

    def __init__(self, nx, ny, na, nm, ni, ns, species_labels, version=None,
                 da=None, ta=None, dm=None, tm=None, di=None, ti=None, rpa=None, rpm=None, ppa=None, ppm=None, rea=None,
                 rem=None, pea=None, pem=None, emist=None, emism=None, elosm=None, edism=None, eradt=None,
                 *args, **kwargs):
        """ Class for holding EIRENE neutral simulation data

        :param file_path:
        :param nx: Number of grid cells in the x direction
        :param ny: Number of grid cells in the y direction
        :param na: Number of atom species
        :param nm: Number of molecule species
        :param ni: Number of ion species
        :param ns: Total number of species in simulation
        :param species_labels: Text descriptions for each species in simulation.
        :param version: Version of EIRENE
        :param da: Atomic Neutral Density
        :param ta: Atomic Neutral Temperature
        :param dm: Molecular Neutral Density
        :param tm: Molecular Neutral Temperature
        :param di: Test Ion Density
        :param ti: Test Ion Temperature
        :param rpa: Atomic Radial Particle Flux
        :param rpm: Molecular Radial Particle Flux
        :param ppa: Atomic Poloidal Particle Flux
        :param ppm: Molecular Poloidal Particle Flux
        :param rea: Atomic Radial Energy Flux
        :param rem: Molecular Radial Energy Flux
        :param pea: Atomic Poloidal Energy Flux
        :param pem: Molecular Poloidal Energy Flux
        :param emist: Total Halpha Emission (including molecules)
        :param emism: Molecular Halpha Emission
        :param elosm: Power loss due to molecules (including dissociation)
        :param edism: Power loss due to molecule dissociation
        :param eradt: Neutral radiated power
        :param debug: Print out debugging information.
        """

        self._nx = nx
        self._ny = ny
        self._na = na
        self._nm = nm
        self._ni = ni
        self._ns = ns

        self._species_labels = species_labels
        self._version = version
        self._da = da
        self._ta = ta
        self._dm = dm
        self._tm = tm
        self._di = di
        self._ti = ti
        self._rpa = rpa
        self._rpm = rpm
        self._ppa = ppa
        self._ppm = ppm
        self._rea = rea
        self._rem = rem
        self._pea = pea
        self._pem = pem
        self._emist = emist
        self._emism = emism
        self._elosm = elosm
        self._edism = edism
        self._eradt = eradt

    @property
    def nx(self):
        """
        Number of grid cells in the poloidal direction

        :rtype: int
        """
        return self._nx

    @property
    def ny(self):
        """
        Number of grid cells in the radial direction

        :rtype: int
        """
        return self._ny

    @property
    def version(self):
        """
        Version of EIRENE

        :rtype: str
        """
        return self._version

    @property
    def na(self):
        """
        Number of atom species

        :rtype: int
        """
        return self._na

    @property
    def nm(self):
        """
        Number of molecule species

        :rtype: int
        """
        return self._nm

    @property
    def ni(self):
        """
        Number of ion species

        :rtype: int
        """
        return self._ni

    @property
    def ns(self):
        """
        Total number of species in simulation

        :rtype: int
        """
        return self._ns

    @property
    def species_labels(self):
        """
        Text descriptions for each species in simulation.

        :rtype: list
        """
        return self._species_labels

    @property
    def da(self):
        """
        Atomic Neutral Density

        :rtype: np.ndarray
        """
        return self._da

    @da.setter
    def da(self, value):
        self._check_dimensions(value, self._na)
        self._da = value

    @property
    def ta(self):
        """
        Atomic Neutral Temperature

        :rtype: np.ndarray
        """
        return self._ta

    @ta.setter
    def ta(self, value):
        self._check_dimensions(value, self._na)
        self._ta = value

    @property
    def dm(self):
        """
        Molecular Neutral Density

        :rtype: np.ndarray
        """
        return self._dm

    @dm.setter
    def dm(self, value):
        self._check_dimensions(value, self._nm)
        self._dm = value

    @property
    def tm(self):
        """
        Molecular Neutral Temperature

        :rtype: np.ndarray
        """
        return self._tm

    @tm.setter
    def tm(self, value):
        self._check_dimensions(value, self._nm)
        self._tm = value

    @property
    def di(self):
        """
        Test Ion Density

        :rtype: np.ndarray
        """
        return self._di

    @di.setter
    def di(self, value):
        self._check_dimensions(value, self._ni)
        self._di = value

    @property
    def ti(self):
        """
        Test Ion Temperature

        :rtype: np.ndarray
        """
        return self._ti

    @ti.setter
    def ti(self, value):
        self._check_dimensions(value, self._ni)
        self._ti = value

    @property
    def rpa(self):
        """
        Atomic Radial Particle Flux

        :rtype: np.ndarray
        """
        return self._rpa

    @rpa.setter
    def rpa(self, value):
        self._check_dimensions(value, self._na)
        self._rpa = value

    @property
    def rpm(self):
        """
        Molecular Radial Particle Flux

        :rtype: np.ndarray
        """
        return self._rpm

    @rpm.setter
    def rpm(self, value):
        self._check_dimensions(value, self._nm)
        self._rpm = value

    @property
    def ppa(self):
        """
        Atomic Poloidal Particle Flux

        :rtype: np.ndarray
        """
        return self._ppa

    @ppa.setter
    def ppa(self, value):
        self._check_dimensions(value, self._na)
        self._ppa = value

    @property
    def ppm(self):
        """
        Molecular Poloidal Particle Flux

        :rtype: np.ndarray
        """
        return self._ppm

    @ppm.setter
    def ppm(self, value):
        self._check_dimensions(value, self._nm)
        self._ppm = value

    @property
    def rea(self):
        """
        Atomic Radial Energy Flux

        :rtype: np.ndarray
        """
        return self._rea

    @rea.setter
    def rea(self, value):
        self._check_dimensions(value, self._na)
        self._rea = value

    @property
    def rem(self):
        """
        Molecular Radial Energy Flux

        :rtype: np.ndarray
        """
        return self._rem

    @rem.setter
    def rem(self, value):
        self._check_dimensions(value, self._nm)
        self._rem = value

    @property
    def pea(self):
        """
        Atomic Poloidal Energy Flux

        :rtype: np.ndarray
        """
        return self._pea

    @pea.setter
    def pea(self, value):
        self._check_dimensions(value, self._na)
        self._pea = value

    @property
    def pem(self):
        """
        Molecular Poloidal Energy Flux

        :rtype: np.ndarray
        """
        return self._pem

    @pem.setter
    def pem(self, value):
        self._check_dimensions(value, self._nm)
        self._pem = value

    @property
    def emist(self):
        """
        Total Halpha Emission (including molecules)

        :rtype: np.ndarray
        """
        return self._emist

    @emist.setter
    def emist(self, value):
        self._check_dimensions(value, 1)
        self._emist = value

    @property
    def emism(self):
        """
        Molecular Halpha Emission

        :rtype: np.ndarray
        """
        return self._emism

    @emism.setter
    def emism(self, value):
        self._check_dimensions(value, 1)
        self._emism = value

    @property
    def elosm(self):
        """
        Power loss due to molecules (including dissociation)

        :rtype: np.ndarray
        """
        return self._elosm

    @elosm.setter
    def elosm(self, value):
        self._check_dimensions(value, self._nm)
        self._elosm = value

    @property
    def edism(self):
        """
        Power loss due to molecule dissociation

        :rtype: np.ndarray
        """
        return self._edism

    @edism.setter
    def edism(self, value):
        self._check_dimensions(value, self._nm)
        self._edism = value

    @property
    def eradt(self):
        """
        Neutral radiated power

        :rtype: np.ndarray
        """
        return self._eradt

    @eradt.setter
    def eradt(self, value):
        self._check_dimensions(value, 1)
        self._eradt = value

    def _check_dimensions(self, data, dim0):
        """
        Checks compatibility of the data array dimension with the species number and grid size.
        :param dim0: size of the 1st dimenion
        :return:
        """

        if not data.shape == (dim0, self._ny, self._nx):
            raise ValueError("Array with shape {0} obtained, but {1} expected".format(data.shape,
                                                                                      (dim0, self._ny, self._nx)))




# Copyright 2016-2021 Euratom
# Copyright 2016-2021 United Kingdom Atomic Energy Authority
# Copyright 2016-2021 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.

import pickle
import numpy as np
from scipy.constants import atomic_mass, electron_mass

# Raysect imports
from raysect.core import translate, Vector3D
from raysect.primitive import Cylinder

# CHERAB core imports
from cherab.core import Plasma, Species, Maxwellian
from cherab.core.math.function import ConstantVector3D
from cherab.core.math.mappers import AxisymmetricMapper, VectorAxisymmetricMapper
from cherab.core.atomic.elements import lookup_isotope, lookup_element

# This SOLPS package imports
from cherab.solps.eirene import Eirene
from .solps_2d_functions import SOLPSFunction2D, SOLPSVectorFunction2D
from .mesh_geometry import SOLPSMesh


class SOLPSSimulation:

    def __init__(self, mesh, species_list):

        # Mesh and species_list cannot be changed after initialisation
        if isinstance(mesh, SOLPSMesh):
            self._mesh = mesh
        else:
            raise ValueError('Argument "mesh" must be a SOLPSMesh instance.')

        if not len(species_list):
            raise ValueError('Argument "species_list" must contain at least one species.')
        self._species_list = tuple(species_list)  # adding additional species is not allowed

        self._initial_setup()

    def _initial_setup(self):

        # Make Mesh Interpolator function for inside/outside mesh test.
        inside_outside_data = np.ones((self._mesh.ny, self._mesh.nx))
        self._inside_mesh = SOLPSFunction2D(self._mesh.vertex_coordinates, self._mesh.triangles,
                                            self._mesh.triangle_to_grid_map, inside_outside_data)

        # Creating a sample SOLPSVectorFunction2D for KDtree to use later
        sample_vector = np.ones((3, self._mesh.ny, self._mesh.nx))
        self._sample_vector_f2d = SOLPSVectorFunction2D(self._mesh.vertex_coordinates, self._mesh.triangles,
                                                        self._mesh.triangle_to_grid_map, sample_vector)

        self._neutral_list = tuple([sp for sp in self._species_list if sp[1] == 0])

        self._electron_temperature = None
        self._electron_temperature_f2d = None
        self._electron_temperature_f3d = None
        self._electron_density = None
        self._electron_density_f2d = None
        self._electron_density_f3d = None
        self._electron_velocities = None
        self._electron_velocities_cylindrical = None
        self._electron_velocities_cylindrical_f2d = None
        self._electron_velocities_cartesian = None
        self._ion_temperature = None
        self._ion_temperature_f2d = None
        self._ion_temperature_f3d = None
        self._neutral_temperature = None
        self._neutral_temperature_f2d = None
        self._neutral_temperature_f3d = None
        self._species_density = None
        self._species_density_f2d = None
        self._species_density_f3d = None
        self._velocities = None
        self._velocities_cylindrical = None
        self._velocities_cylindrical_f2d = None
        self._velocities_cartesian = None
        self._halpha_mol_radiation = None
        self._halpha_mol_radiation_f2d = None
        self._halpha_mol_radiation_f3d = None
        self._halpha_total_radiation = None
        self._halpha_total_radiation_f2d = None
        self._halpha_total_radiation_f3d = None
        self._total_radiation = None
        self._total_radiation_f2d = None
        self._total_radiation_f3d = None
        self._b_field = None
        self._b_field_cylindrical = None
        self._b_field_cylindrical_f2d = None
        self._b_field_cartesian = None
        self._eirene_model = None
        self._b2_model = None
        self._eirene = None

    @property
    def mesh(self):
        """
        SOLPSMesh instance.
        :return:
        """
        return self._mesh

    @property
    def species_list(self):
        """
        Tuple of species elements in the form (species name, charge).
        :return:
        """
        return self._species_list

    @property
    def neutral_list(self):
        """
        Tuple of species elements in the form (species name, charge).
        :return:
        """
        return self._neutral_list

    @property
    def electron_temperature(self):
        """
        Simulated electron temperatures at each mesh cell.
        Array of shape (ny, nx).
        :return:
        """
        return self._electron_temperature

    @property
    def electron_temperature_f2d(self):
        """
        Function2D interpolator for electron temperature.
        Returns electron temperature at a given point (R, Z).
        :return:
        """
        return self._electron_temperature_f2d

    @property
    def electron_temperature_f3d(self):
        """
        Function3D interpolator for electron temperature.
        Returns electron temperature at a given point (x, y, z).
        :return:
        """
        return self._electron_temperature_f3d

    @electron_temperature.setter
    def electron_temperature(self, value):
        value = np.array(value, dtype=np.float64, copy=False)
        _check_shape("electron_temperature", value, (self.mesh.ny, self.mesh.nx))
        self._electron_temperature = value
        self._electron_temperature_f2d = SOLPSFunction2D.instance(self._inside_mesh, value)
        self._electron_temperature_f3d = AxisymmetricMapper(self._electron_temperature_f2d)

    @property
    def ion_temperature(self):
        """
        Simulated ion temperatures at each mesh cell.
        Array of shape (ny, nx).
        :return:
        """
        return self._ion_temperature

    @property
    def ion_temperature_f2d(self):
        """
        Function2D interpolator for ion temperature.
        Returns ion temperature at a given point (R, Z).
        :return:
        """
        return self._ion_temperature_f2d

    @property
    def ion_temperature_f3d(self):
        """
        Function3D interpolator for ion temperature.
        Returns ion temperature at a given point (x, y, z).
        :return:
        """
        return self._ion_temperature_f3d

    @ion_temperature.setter
    def ion_temperature(self, value):
        value = np.array(value, dtype=np.float64, copy=False)
        _check_shape("ion_temperature", value, (self.mesh.ny, self.mesh.nx))
        self._ion_temperature = value
        self._ion_temperature_f2d = SOLPSFunction2D.instance(self._inside_mesh, value)
        self._ion_temperature_f3d = AxisymmetricMapper(self._ion_temperature_f2d)

    @property
    def neutral_temperature(self):
        """
        Simulated neutral atom (effective) temperatures at each mesh cell.
        Array of shape (na, ny, nx).
        :return:
        """
        return self._neutral_temperature

    @property
    def neutral_temperature_f2d(self):
        """
        Dictionary of Function2D interpolators for neutral atom (effective) temperatures.
        Accessed by neutral atom index or neutral_list elements.
        E.g., neutral_temperature_f2d[0] or neutral_temperature_f2d[('deuterium', 0))].
        Each entry returns neutral atom (effective) temperature at a given point (R, Z).
        :return:
        """
        return self._neutral_temperature_f2d

    @property
    def neutral_temperature_f3d(self):
        """
        Dictionary of Function3D interpolators for neutral atom (effective) temperatures.
        Accessed by neutral atom index or neutral_list elements.
        E.g., neutral_temperature_f3d[0] or neutral_temperature_f3d[('deuterium', 0))].
        Each entry returns neutral atom (effective) temperature at a given point (x, y, z).
        :return:
        """
        return self._neutral_temperature_f3d

    @neutral_temperature.setter
    def neutral_temperature(self, value):
        value = np.array(value, dtype=np.float64, copy=False)
        _check_shape("neutral_temperature", value, (len(self._neutral_list), self.mesh.ny, self.mesh.nx))
        self._neutral_temperature = value
        self._neutral_temperature_f2d = {}
        self._neutral_temperature_f3d = {}
        for k, sp in enumerate(self._neutral_list):
            self._neutral_temperature_f2d[k] = SOLPSFunction2D.instance(self._inside_mesh, value[k])
            self._neutral_temperature_f2d[sp] = self._neutral_temperature_f2d[k]
            self._neutral_temperature_f3d[k] = AxisymmetricMapper(self._neutral_temperature_f2d[k])
            self._neutral_temperature_f3d[sp] = self._neutral_temperature_f3d[k]

    @property
    def electron_density(self):
        """
        Simulated electron densities at each mesh cell.
        Array of shape (ny, nx)
        :return:
        """
        return self._electron_density

    @property
    def electron_density_f2d(self):
        """
        Function2D interpolator for electron density.
        Returns electron density at a given point (R, Z).
        :return:
        """
        return self._electron_density_f2d

    @property
    def electron_density_f3d(self):
        """
        Function3D interpolator for electron density.
        Returns electron density at a given point (x, y, z).
        :return:
        """
        return self._electron_density_f3d

    @electron_density.setter
    def electron_density(self, value):
        value = np.array(value, dtype=np.float64, copy=False)
        _check_shape("electron_density", value, (self.mesh.ny, self.mesh.nx))
        self._electron_density = value
        self._electron_density_f2d = SOLPSFunction2D.instance(self._inside_mesh, value)
        self._electron_density_f3d = AxisymmetricMapper(self._electron_density_f2d)

    @property
    def electron_velocities(self):
        """
        Electron velocities in poloidal coordinates (e_pol, e_rad, e_tor) at each mesh cell.
        Array of shape (3, ny, nx):
        [0, :, :] - poloidal, [1, :, :] - radial, [2, :, :] - toroidal.
        :return:
        """
        return self._electron_velocities

    @electron_velocities.setter
    def electron_velocities(self, value):
        value = np.array(value, dtype=np.float64, copy=False)
        _check_shape("electron_velocities", value, (3, self.mesh.ny, self.mesh.nx))

        # Converting to cylindrical coordinates
        velocities_cylindrical = np.zeros(value.shape)
        velocities_cylindrical[1] = -value[2]
        velocities_cylindrical[[0, 2]] = self.mesh.to_cartesian(value[:2])

        self._electron_velocities = value
        self._electron_velocities_cylindrical = velocities_cylindrical
        self._electron_velocities_cylindrical_f2d = SOLPSVectorFunction2D.instance(self._sample_vector_f2d, velocities_cylindrical)
        self._electron_velocities_cartesian = VectorAxisymmetricMapper(self._electron_velocities_cylindrical_f2d)

    @property
    def electron_velocities_cylindrical(self):
        """
        Electron velocities in cylindrical coordinates (R, phi, Z) at each mesh cell.
        Array of shape (3, ny, nx): [0, :, :] - R, [1, :, :] - phi, [2, :, :] - Z.
        :return:
        """
        return self._electron_velocities_cylindrical

    @electron_velocities_cylindrical.setter
    def electron_velocities_cylindrical(self, value):
        value = np.array(value, dtype=np.float64, copy=False)
        _check_shape("electron_velocities_cylindrical", value, (3, self.mesh.ny, self.mesh.nx))

        # Converting to poloidal coordinates
        velocities = np.zeros(value.shape)
        velocities[2] = -value[1]
        velocities[:2] = self.mesh.to_poloidal(value[[0, 2]])

        self._electron_velocities_cylindrical = value
        self._electron_velocities = velocities
        self._electron_velocities_cylindrical_f2d = SOLPSVectorFunction2D.instance(self._sample_vector_f2d, value)
        self._electron_velocities_cartesian = VectorAxisymmetricMapper(self._electron_velocities_cylindrical_f2d)

    @property
    def electron_velocities_cylindrical_f2d(self):
        """
        VectorFunction2D interpolator for electron velocities in cylindrical coordinates.
        Returns a vector of electron velocity at a given point (R, Z).
        :return:
        """
        return self._electron_velocities_cylindrical_f2d

    @property
    def electron_velocities_cartesian(self):
        """
        VectorFunction3D interpolator for electron velocities in Cartesian coordinates.
        Returns a vector of electron velocity at a given point (x, y, z).
        :return:
        """
        return self._electron_velocities_cartesian

    @property
    def species_density(self):
        """
        Simulated species densities at each mesh cell.
        Array of shape (ns, ny, nx).
        :return:
        """
        return self._species_density

    @property
    def species_density_f2d(self):
        """
        Dictionary of Function2D interpolators for species densities.
        Accessed by species index or species_list elements.
        E.g., species_density_f2d[1] or species_density_f2d[('deuterium', 1))].
        Each entry returns species density at a given point (R, Z).
        :return:
        """
        return self._species_density_f2d

    @property
    def species_density_f3d(self):
        """
        Dictionary of Function3D interpolators for species densities.
        Accessed by species index or species_list elements.
        E.g., species_density_f3d[1] or species_density_f3d[('deuterium', 1))].
        Each entry returns species density at a given point (x, y, z).
        :return:
        """
        return self._species_density_f3d

    @species_density.setter
    def species_density(self, value):
        value = np.array(value, dtype=np.float64, copy=False)
        _check_shape("species_density", value, (len(self._species_list), self.mesh.ny, self.mesh.nx))
        self._species_density = value
        self._species_density_f2d = {}
        self._species_density_f3d = {}
        for k, sp in enumerate(self._species_list):
            self._species_density_f2d[k] = SOLPSFunction2D.instance(self._inside_mesh, value[k])
            self._species_density_f2d[sp] = self._species_density_f2d[k]
            self._species_density_f3d[k] = AxisymmetricMapper(self._species_density_f2d[k])
            self._species_density_f3d[sp] = self._species_density_f3d[k]

    @property
    def velocities(self):
        """
        Species velocities in poloidal coordinates (e_pol, e_rad, e_tor) at each mesh cell.
        Array of shape (ns, 3, ny, nx):
        [:, 0, :, :] - poloidal, [:, 1, :, :] - radial, [:, 2, :, :] - toroidal.
        :return:
        """
        return self._velocities

    @velocities.setter
    def velocities(self, value):
        value = np.array(value, dtype=np.float64, copy=False)
        _check_shape("velocities", value, (len(self.species_list), 3, self.mesh.ny, self.mesh.nx))

        # Converting to cylindrical coordinates
        velocities_cylindrical = np.zeros(value.shape)
        velocities_cylindrical[:, 1] = -value[:, 2]
        for k in range(value.shape[0]):
            velocities_cylindrical[k, [0, 2]] = self.mesh.to_cartesian(value[k, :2])

        self._velocities = value
        self._velocities_cylindrical = velocities_cylindrical
        self._velocities_cylindrical_f2d = {}
        self._velocities_cartesian = {}
        for k, sp in enumerate(self._species_list):
            self._velocities_cylindrical_f2d[k] = SOLPSVectorFunction2D.instance(self._sample_vector_f2d, velocities_cylindrical[k])
            self._velocities_cylindrical_f2d[sp] = self._velocities_cylindrical_f2d[k]
            self._velocities_cartesian[k] = VectorAxisymmetricMapper(self._velocities_cylindrical_f2d[k])
            self._velocities_cartesian[sp] = self._velocities_cartesian[k]

    @property
    def velocities_cylindrical(self):
        """
        Species velocities in cylindrical coordinates (R, phi, Z) at each mesh cell.
        Array of shape (ns, 3, ny, nx): [:, 0, :, :] - R, [:, 1, :, :] - phi, [:, 2, :, :] - Z.
        :return:
        """
        return self._velocities_cylindrical

    @velocities_cylindrical.setter
    def velocities_cylindrical(self, value):
        value = np.array(value, dtype=np.float64, copy=False)
        _check_shape("velocities_cylindrical", value, (len(self.species_list), 3, self.mesh.ny, self.mesh.nx))

        # Converting to poloidal coordinates
        velocities = np.zeros(value.shape)
        velocities[:, 2] = -value[:, 1]
        for k in range(value.shape[0]):
            velocities[k, :2] = self.mesh.to_poloidal(value[k, [0, 2]])

        self._velocities_cylindrical = value
        self._velocities = velocities
        self._velocities_cylindrical_f2d = {}
        self._velocities_cartesian = {}
        for k, sp in enumerate(self._species_list):
            self._velocities_cylindrical_f2d[k] = SOLPSVectorFunction2D.instance(self._sample_vector_f2d, value[k])
            self._velocities_cylindrical_f2d[sp] = self._velocities_cylindrical_f2d[k]
            self._velocities_cartesian[k] = VectorAxisymmetricMapper(self._velocities_cylindrical_f2d[k])
            self._velocities_cartesian[sp] = self._velocities_cartesian[k]

    @property
    def velocities_cylindrical_f2d(self):
        """
        Dictionary of VectorFunction2D interpolators for species velocities
        in cylindrical coordinates.
        Accessed by species index or species_list elements.
        E.g., velocities_cylindrical_f2d[1] or velocities_cylindrical_f2d[('deuterium', 1))].
        Each entry returns a vector of species velocity at a given point (R, Z).
        :return:
        """
        return self._velocities_cylindrical_f2d

    @property
    def velocities_cartesian(self):
        """
        Dictionary of VectorFunction3D interpolators for species velocities
        in Cartesian coordinates.
        Accessed by species index or species_list elements.
        E.g., velocities_cartesian[1] or velocities_cartesian[('deuterium', 1))].
        Each entry returns a vector of species velocity at a given point (x, y, z).
        :return:
        """
        return self._velocities_cartesian

    @property
    def inside_mesh(self):
        """
        Function2D for testing if point p is inside the simulation mesh.
        """
        return self._inside_mesh

    @property
    def inside_volume_mesh(self):
        """
        Function3D for testing if point p is inside the simulation mesh.
        """
        return AxisymmetricMapper(self._inside_mesh)

    @property
    def total_radiation(self):
        """
        Total radiation at each mesh cell.
        Array of shape (ny, nx).

        This is not calculated from the CHERAB emission models, instead it comes from the SOLPS output data.
        Is calculated from the sum of all integrated line emission and all Bremmstrahlung.
        The signals used are 'RQRAD' and 'RQBRM'. Final output is in W m-3.
        """

        return self._total_radiation

    @property
    def total_radiation_f2d(self):
        """
        Function2D interpolator for total radiation.
        Returns total radiation at a given point (R, Z).
        """

        return self._total_radiation_f2d

    @property
    def total_radiation_f3d(self):
        """
        Function3D interpolator for total radiation.
        Returns total radiation at a given point (x, y, z).
        """

        return self._total_radiation_f3d

    @total_radiation.setter
    def total_radiation(self, value):
        value = np.array(value, dtype=np.float64, copy=False)
        _check_shape("total_radiation", value, (self.mesh.ny, self.mesh.nx))
        self._total_radiation = value
        self._total_radiation_f2d = SOLPSFunction2D.instance(self._inside_mesh, value)
        self._total_radiation_f3d = AxisymmetricMapper(self._total_radiation_f2d)

    @property
    def halpha_mol_radiation(self):
        """
        Molecular H-alpha radiation at each mesh cell.
        Array of shape (ny, nx).

        This is calculated with a built-in EIRENE collissional radiative model (CRM).
        Final output is in W m-3.
        """

        return self._halpha_mol_radiation

    @property
    def halpha_mol_radiation_f2d(self):
        """
        Function2D interpolator for molecular H-alpha radiation.
        Returns total radiation at a given point (R, Z).
        """

        return self._halpha_mol_radiation_f2d

    @property
    def halpha_mol_radiation_f3d(self):
        """
        Function3D interpolator for molecular H-alpha radiation.
        Returns total radiation at a given point (x, y, z).
        """

        return self._halpha_mol_radiation_f3d

    @halpha_mol_radiation.setter
    def halpha_mol_radiation(self, value):
        value = np.array(value, dtype=np.float64, copy=False)
        _check_shape("halpha_mol_radiation", value, (self.mesh.ny, self.mesh.nx))
        self._halpha_mol_radiation = value
        self._halpha_mol_radiation_f2d = SOLPSFunction2D.instance(self._inside_mesh, value)
        self._halpha_mol_radiation_f3d = AxisymmetricMapper(self._halpha_mol_radiation_f2d)

    @property
    def halpha_total_radiation(self):
        """
        Total (molecular + atomic) H-alpha radiation at each mesh cell.
        Array of shape (ny, nx).

        This is calculated with a built-in EIRENE collissional radiative model (CRM).
        Final output is in W m-3.
        """

        return self._halpha_total_radiation

    @property
    def halpha_total_radiation_f2d(self):
        """
        Function2D interpolator for total H-alpha radiation.
        Returns total radiation at a given point (R, Z).
        """

        return self._halpha_total_radiation_f2d

    @property
    def halpha_total_radiation_f3d(self):
        """
        Function3D interpolator for total H-alpha radiation.
        Returns total radiation at a given point (x, y, z).
        """

        return self._halpha_total_radiation_f3d

    @halpha_total_radiation.setter
    def halpha_total_radiation(self, value):
        value = np.array(value, dtype=np.float64, copy=False)
        _check_shape("halpha_total_radiation", value, (self.mesh.ny, self.mesh.nx))
        self._halpha_total_radiation = value
        self._halpha_total_radiation_f2d = SOLPSFunction2D.instance(self._inside_mesh, value)
        self._halpha_total_radiation_f3d = AxisymmetricMapper(self._halpha_total_radiation_f2d)

    @property
    def b_field(self):
        """
        Magnetic B field in poloidal coordinates (e_pol, e_rad, e_tor) at each mesh cell.
        Array of shape (3, ny, nx): [0, :, :] - poloidal, [1, :, :] - radial, [2, :, :] - toroidal.
        """

        return self._b_field

    @b_field.setter
    def b_field(self, value):
        value = np.array(value, dtype=np.float64, copy=False)
        _check_shape("b_field", value, (3, self.mesh.ny, self.mesh.nx))

        # Converting to cylindrical system
        b_field_cylindrical = np.zeros(value.shape)
        b_field_cylindrical[1] = -value[2]
        b_field_cylindrical[[0, 2]] = self.mesh.to_cartesian(value[:2])

        self._b_field_cylindrical = b_field_cylindrical
        self._b_field = value
        self._b_field_cylindrical_f2d = SOLPSVectorFunction2D.instance(self._sample_vector_f2d, b_field_cylindrical)
        self._b_field_cartesian = VectorAxisymmetricMapper(self._b_field_cylindrical_f2d)

    @property
    def b_field_cylindrical(self):
        """
        Magnetic B field in poloidal coordinates (R, phi, Z) at each mesh cell.
        Array of shape (3, ny, nx): [0, :, :] - R, [1, :, :] - phi, [2, :, :] - Z.
        """

        return self._b_field_cylindrical

    @property
    def b_field_cylindrical_f2d(self):
        """
        VectorFunction2D interpolator for magnetic B field in cylindrical coordinates.
        Returns a vector of magnetic field at a given point (R, Z).
        """

        return self._b_field_cylindrical_f2d

    @property
    def b_field_cartesian(self):
        """
        VectorFunction3D interpolator for magnetic B field in Cartesian coordinates.
        Returns a vector of magnetic field at a given point (x, y, z).
        """

        return self._b_field_cartesian

    @b_field_cylindrical.setter
    def b_field_cylindrical(self, value):
        value = np.array(value, dtype=np.float64, copy=False)
        _check_shape("b_field_cylindrical", value, (3, self.mesh.ny, self.mesh.nx))

        # Converting to poloidal system
        b_field = np.zeros(value.shape)
        b_field[2] = -value[1]
        b_field[:2] = self.mesh.to_poloidal(value[[0, 2]])

        self._b_field = b_field
        self._b_field_cylindrical = value
        self._b_field_cylindrical_f2d = SOLPSVectorFunction2D.instance(self._sample_vector_f2d, value)
        self._b_field_cartesian = VectorAxisymmetricMapper(self._b_field_cylindrical_f2d)

    @property
    def eirene_simulation(self):
        """
        Data from an underlying EIRENE simulation.

        :rtype: Eirene
        """

        return self._eirene

    @eirene_simulation.setter
    def eirene_simulation(self, value):
        if not isinstance(value, Eirene):
            raise ValueError('Attribute "eirene_simulation" must be an Eirene instance.')

        self._eirene = value

    def __getstate__(self):
        state = {
            'mesh': self._mesh,
            'species_list': self._species_list,
            'electron_temperature': self._electron_temperature,
            'ion_temperature': self._ion_temperature,
            'neutral_temperature': self._neutral_temperature,
            'electron_density': self._electron_density,
            'species_density': self._species_density,
            'electron_velocities_cylindrical': self._electron_velocities_cylindrical,
            'velocities_cylindrical': self._velocities_cylindrical,
            'total_radiation': self._total_radiation,
            'halpha_mol_radiation': self._halpha_mol_radiation,
            'halpha_total_radiation': self._halpha_total_radiation,
            'b_field_cylindrical': self._b_field_cylindrical,
            'eirene_model': self._eirene_model,
            'b2_model': self._b2_model,
            'eirene': self._eirene
        }
        return state

    def __setstate__(self, state):
        self._mesh = state['mesh']
        self._species_list = state['species_list']
        self._initial_setup()
        if state['electron_temperature'] is not None:
            self.electron_temperature = state['electron_temperature']  # will create _f2d() and _f3d()
        if state['ion_temperature'] is not None:
            self.ion_temperature = state['ion_temperature']
        if state['neutral_temperature'] is not None:
            self.neutral_temperature = state['neutral_temperature']
        if state['electron_density'] is not None:
            self.electron_density = state['electron_density']
        if state['species_density'] is not None:
            self.species_density = state['species_density']
        if state['electron_velocities_cylindrical'] is not None:
            self.electron_velocities_cylindrical = state['electron_velocities_cylindrical']
        if state['velocities_cylindrical'] is not None:
            self.velocities_cylindrical = state['velocities_cylindrical']
        if state['total_radiation'] is not None:
            self.total_radiation = state['total_radiation']
        if state['halpha_mol_radiation'] is not None:
            self.halpha_mol_radiation = state['halpha_mol_radiation']
        if state['halpha_total_radiation'] is not None:
            self.halpha_total_radiation = state['halpha_total_radiation']
        if state['b_field_cylindrical'] is not None:
            self.b_field_cylindrical = state['b_field_cylindrical']
        self._eirene_model = state['eirene_model']
        self._b2_model = state['b2_model']
        self._eirene = state['eirene']

    def save(self, filename):
        """
        Saves SOLPSSimulation to file.
        """
        with open(filename, 'wb') as file_handle:
            pickle.dump(self, file_handle)

    @classmethod
    def load(cls, filename):
        """
        Loads SOLPSSimulation from file.
        """
        with open(filename, 'rb') as file_handle:
            sim = pickle.load(file_handle)

        return sim

    def create_plasma(self, parent=None, transform=None, name=None):
        """
        Make a CHERAB plasma object from this SOLPS simulation.

        :param Node parent: The plasma's parent node in the scenegraph, e.g. a World object.
        :param AffineMatrix3D transform: Affine matrix describing the location and orientation
        of the plasma in the world.
        :param str name: User friendly name for this plasma (default = "SOLPS Plasma").
        :rtype: Plasma
        """

        # Checking if the minimal required data is available to create a plasma object
        if self.electron_density_f3d is None:
            raise RuntimeError("Unable to create plasma object: electron density is not set.")
        if self.electron_temperature_f3d is None:
            raise RuntimeError("Unable to create plasma object: electron temperature is not set.")
        if self.species_density_f3d is None:
            raise RuntimeError("Unable to create plasma object: species density is not set.")
        if self.ion_temperature_f3d is None:
            raise RuntimeError("Unable to create plasma object: ion temperature is not set.")

        mesh = self.mesh
        name = name or "SOLPS Plasma"
        plasma = Plasma(parent=parent, transform=transform, name=name)
        radius = mesh.mesh_extent['maxr']
        height = mesh.mesh_extent['maxz'] - mesh.mesh_extent['minz']
        plasma.geometry = Cylinder(radius, height)
        plasma.geometry_transform = translate(0, 0, mesh.mesh_extent['minz'])

        if self.b_field_cartesian is None:
            print('Warning! No magnetic field data available for this simulation.')
        else:
            plasma.b_field = self.b_field_cartesian

        # Create electron species
        if self.electron_velocities_cartesian is None:
            print('Warning! No electron velocity data available for this simulation.')
            electron_velocity = ConstantVector3D(Vector3D(0, 0, 0))
        else:
            electron_velocity = self.electron_velocities_cartesian
        plasma.electron_distribution = Maxwellian(self.electron_density_f3d, self.electron_temperature_f3d, electron_velocity, electron_mass)

        if self.velocities_cartesian is None:
            print('Warning! No species velocities data available for this simulation.')

        if self.neutral_temperature_f3d is None:
            print('Warning! No neutral atom temperature data available for this simulation.')

        neutral_i = 0  # neutrals count
        for k, sp in enumerate(self.species_list):

            try:
                species_type = lookup_element(sp[0])
            except ValueError:
                species_type = lookup_isotope(sp[0])

            charge = sp[1]

            # Create the velocity vector lookup function
            if self.velocities_cartesian is not None:
                velocity = self.velocities_cartesian[k]
            else:
                velocity = ConstantVector3D(Vector3D(0, 0, 0))

            if charge or self.neutral_temperature is None:  # ions or neutral atoms (neutral temperature is not available)
                distribution = Maxwellian(self.species_density_f3d[k], self.ion_temperature_f3d, velocity,
                                          species_type.atomic_weight * atomic_mass)

            else:  # neutral atoms with neutral temperature
                distribution = Maxwellian(self.species_density_f3d[k], self._neutral_temperature_f3d[neutral_i], velocity,
                                          species_type.atomic_weight * atomic_mass)
                neutral_i += 1

            plasma.composition.add(Species(species_type, charge, distribution))

        return plasma


def _check_shape(name, value, shape):
    if value.shape != shape:
        raise ValueError('Shape of "{0}": {1} mismatch the shape of SOLPS grid: {2}.'.format(name, value.shape, shape))


def prefer_element(isotope):
    """
    Return Element instance, if the element of this isotope has the same mass number.
    """
    el_mass_number = int(round(isotope.element.atomic_weight))
    if el_mass_number == isotope.mass_number:
        return isotope.element

    return isotope


def b2_flux_to_velocity(sim, poloidal_flux, radial_flux, parallel_velocity):
    """
    Calculates velocities of all species at cell centres using B2 particle fluxes defined
    at cell faces.

    :param SOLPSSimulation sim: SOLPSSimulation instance.
    :param ndarray poloidal_flux: Poloidal flux of species in s-1. Must be a 3 dimensional array of
                                  shape (num_species, mesh.ny, mesh.nx).
    :param ndarray radial_flux: Radial flux of species in s-1. Must be a 3 dimensional array of
                                shape (num_species, mesh.ny, mesh.nx).
    :param ndarray parallel_velocity: Parallel velocity of species in m/s. Must be a 3 dimensional
                                      array of shape (num_species, mesh.ny, mesh.nx).
                                      Parallel velocity is a velocity projection on magnetic
                                      field direction.

    :return ndarray: Velocities of all species in cylindrical coordinates.
                     Array of shape (num_species, 3, mesh.ny, mesh.nx).
    """
    if sim.b_field_cylindrical is None:
        raise RuntimeError('Attribute "b_field_cylindrical" is not set.')
    if sim.species_density is None:
        raise RuntimeError('Attribute "species_density" is not set.')

    mesh = sim.mesh
    b = sim.b_field_cylindrical

    poloidal_flux = np.array(poloidal_flux, dtype=np.float64, copy=False)
    radial_flux = np.array(radial_flux, dtype=np.float64, copy=False)
    parallel_velocity = np.array(parallel_velocity, dtype=np.float64, copy=False)

    nx = mesh.nx  # poloidal
    ny = mesh.ny  # radial
    ns = len(sim.species_list)  # number of species

    _check_shape('poloidal_flux', poloidal_flux, (ns, ny, nx))
    _check_shape('radial_flux', radial_flux, (ns, ny, nx))
    _check_shape('parallel_velocity', parallel_velocity, (ns, ny, nx))

    poloidal_area = mesh.poloidal_area
    radial_area = mesh.radial_area
    leftix = mesh.neighbix[0]  # poloidal prev.
    leftiy = mesh.neighbiy[0]
    bottomix = mesh.neighbix[1]  # radial prev.
    bottomiy = mesh.neighbiy[1]
    rightix = mesh.neighbix[2]   # poloidal next.
    rightiy = mesh.neighbiy[2]
    topix = mesh.neighbix[3]  # radial next.
    topiy = mesh.neighbiy[3]

    # Converting s-1 --> m-2 s-1
    poloidal_flux = np.divide(poloidal_flux, poloidal_area, out=np.zeros_like(poloidal_flux), where=poloidal_area > 0)
    radial_flux = np.divide(radial_flux, radial_area, out=np.zeros_like(radial_flux), where=radial_area > 0)

    # Obtaining left velocity
    dens_neighb = sim.species_density[:, leftiy, leftix]  # density in the left neighbouring cell
    has_neighbour = ((leftix > -1) * (leftiy > -1))  # check if has left neighbour
    neg_flux = (poloidal_flux < 0) * (sim.species_density > 0)  # will use density in this cell if flux is negative
    pos_flux = (poloidal_flux > 0) * (dens_neighb > 0) * has_neighbour  # will use density in neighbouring cell if flux is positive
    velocity_left = np.divide(poloidal_flux, sim.species_density, out=np.zeros((ns, ny, nx)), where=neg_flux)
    velocity_left = np.divide(poloidal_flux, dens_neighb, out=velocity_left, where=pos_flux)
    poloidal_face_normal = mesh.radial_basis_vector[[1, 0]]
    poloidal_face_normal[1] *= -1
    velocity_left = velocity_left[:, None] * poloidal_face_normal  # to vector in RZ

    # Obtaining bottom velocity
    dens_neighb = sim.species_density[:, bottomiy, bottomix]
    has_neighbour = ((bottomix > -1) * (bottomiy > -1))
    neg_flux = (radial_flux < 0) * (sim.species_density > 0)
    pos_flux = (poloidal_flux > 0) * (dens_neighb > 0) * has_neighbour
    velocity_bottom = np.divide(radial_flux, sim.species_density, out=np.zeros((ns, ny, nx)), where=neg_flux)
    velocity_bottom = np.divide(radial_flux, dens_neighb, out=velocity_bottom, where=pos_flux)
    radial_face_normal = mesh.poloidal_basis_vector[[1, 0]]
    radial_face_normal[0] *= -1
    velocity_bottom = velocity_bottom[:, None] * radial_face_normal  # to RZ

    # Obtaining right and top velocities
    velocity_right = velocity_left[:, :, rightiy, rightix]
    velocity_right[:, :, (rightix < 0) + (rightiy < 0)] = 0

    velocity_top = velocity_bottom[:, :, topiy, topix]
    velocity_top[:, :, (topix < 0) + (topiy < 0)] = 0

    vcyl = np.zeros((ns, 3, ny, nx))  # velocities in cylindrical coordinates

    # Projection of velocity on RZ-plane
    vcyl[:, [0, 2]] = 0.25 * (velocity_bottom + velocity_left + velocity_top + velocity_right)

    # Obtaining toroidal velocity
    bmagn = np.sqrt((b * b).sum(0))
    vcyl[:, 1] = (parallel_velocity * bmagn - vcyl[:, 0] * b[0] - vcyl[:, 2] * b[2]) / b[1]

    return vcyl


def eirene_flux_to_velocity(sim, poloidal_flux, radial_flux, parallel_velocity):
    """
    Calculates velocities of neutral atoms using Eirene particle fluxes defined at cell centre.

    :param SOLPSSimulation sim: SOLPSSimulation instance.
    :param ndarray poloidal_flux: Poloidal flux of atoms in m-2 s-1. Must be a 3 dimensional array of
                                  shape (num_atoms, mesh.ny, mesh.nx).
    :param ndarray radial_flux: Radial flux of atoms in m-2 s-1. Must be a 3 dimensional array of
                                shape (num_atoms, mesh.ny, mesh.nx).
    :param ndarray parallel_velocity: Parallel velocity of atoms in m/s. Must be a 3 dimensional
                                      array of shape (num_atoms, mesh.ny, mesh.nx).
                                      Parallel velocity is a velocity projection on magnetic
                                      field direction.

    :return ndarray: Velocities of neutral atoms in cylindrical coordinates.
                     Array of shape (num_atoms, 3, mesh.ny, mesh.nx).
    """
    if sim.b_field_cylindrical is None:
        raise RuntimeError('Attribute "b_field_cylindrical" is not set.')
    if sim.species_density is None:
        raise RuntimeError('Attribute "species_density" is not set.')

    mesh = sim.mesh
    b = sim.b_field_cylindrical

    poloidal_flux = np.array(poloidal_flux, dtype=np.float64, copy=False)
    radial_flux = np.array(radial_flux, dtype=np.float64, copy=False)
    parallel_velocity = np.array(parallel_velocity, dtype=np.float64, copy=False)

    nx = mesh.nx  # poloidal
    ny = mesh.ny  # radial
    ns = len(sim.neutral_list)  # number of neutral atoms

    _check_shape('poloidal_flux', poloidal_flux, (ns, ny, nx))
    _check_shape('radial_flux', radial_flux, (ns, ny, nx))
    _check_shape('parallel_velocity', parallel_velocity, (ns, ny, nx))

    neutral_indx = [k for k, sp in enumerate(sim.species_list) if sp[1] == 0]
    density = sim.species_density[neutral_indx, :]

    # Obtaining velocity
    poloidal_velocity = np.divide(poloidal_flux, density, out=np.zeros_like(density), where=(density > 0))
    radial_velocity = np.divide(radial_flux, density, out=np.zeros_like(density), where=(density > 0))

    vcyl = np.zeros((ns, 3, ny, nx))  # velocities in cylindrical coordinates

    # Projection of velocity on RZ-plane
    vcyl[:, [0, 2]] = (poloidal_velocity[:, None] * mesh.poloidal_basis_vector + radial_velocity[:, None] * mesh.radial_basis_vector)

    # Obtaining toroidal velocity
    bmagn = np.sqrt((b * b).sum(0))
    vcyl[:, 1] = (parallel_velocity * bmagn - vcyl[:, 0] * b[0] - vcyl[:, 2] * b[2]) / b[1]

    return vcyl



# Copyright 2016-2018 Euratom
# Copyright 2016-2018 United Kingdom Atomic Energy Authority
# Copyright 2016-2018 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.

import re
import numpy as np


def read_unlabelled_block44(file_handle, ns, nx, ny):
    """ Read standard block in EIRENE code output file 'fort.44'

    This function supports blocks in fort.44 format versions which do
    not contain a header line. Versions up to 20130210 should work with
    this function.

    :param file_handle: A python core file handle object as a result of a
      call to open('./fort.44').
    :param int ns: total number of species
    :param int nx: number of grid poloidal cells
    :param int ny: number of grid radial cells
    :return: ndarray of data with shape [ns, ny, nx]
    """
    data = []
    npoints = ns * nx * ny
    while len(data) < npoints:
        line = file_handle.readline().split()
        data.extend(line)
    data = np.asarray(data, dtype=float).reshape((ns, ny, nx))
    return data


def read_labelled_block44(file_handle):
    """ Read standard block in EIRENE code output file 'fort.44'

    This function supports blocks in fort.44 format versions which have
    a header line beginning with `*eirene` and containing the name of
    the variable and the number of items. Versions later than 2016
    should work with this function.

    :param file_handle: A python core file handle object as a result of a
      call to open('./fort.44').
    :return: a dictionary {name: data} where name is a string and data is
      a 1D array of the data.
    """
    header = file_handle.readline()
    if not header:
        raise EOFError()
    # Treat blank lines between blocks as null data.
    if header == "\n":
        return {}
    # Remove whitespace in indexed labels.
    pattern = re.compile(r"\(\s+(\d+)\)")
    header = pattern.sub(r"(\1)", header)
    header = header.split()
    # Deal with unlabelled printing of nlim, nsts, nstra.
    if len(header) == 3:
        return {'extra_dims': np.asarray(header, dtype=int)}
    if " ".join(header[:3]) != "*eirene data field":
        raise ValueError("fort.44 block format is not supported.")
    name = header[3]
    npoints = int(header[6])
    data = []
    while len(data) < npoints:
        file_pos = file_handle.tell()
        line = file_handle.readline().split()
        # A bug(?) in 20170328 means strata-resolved quantities are one value short.
        if line[0] == "*eirene":
            data.extend("0")
            file_handle.seek(file_pos)
            break
        # Some variables include species labels on the first line instead of numbers.
        try:
            float(line[0])
        except ValueError:
            continue
        data.extend(line)
    data = np.asarray(data, dtype=float)
    return {name: data}



# cython: language_level=3

# Copyright 2016-2021 Euratom
# Copyright 2016-2021 United Kingdom Atomic Energy Authority
# Copyright 2016-2021 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.

from raysect.optical cimport Spectrum, Point3D, Vector3D
from cherab.core cimport Plasma, AtomicData
from cherab.core.model.lineshape cimport GaussianLine, LineShapeModel
from cherab.core.utility.constants cimport RECIP_4_PI
from cherab.core.math.mappers import AxisymmetricMapper

from cherab.solps cimport SOLPSFunction2D


cdef class SOLPSLineEmitter(PlasmaModel):
    """
    Line emitter that uses a custom `SOLPSFunction2D` function for radiation power density.

    :param SOLPSFunction2D radiation_function: The power density function defined on the SOLPS mesh.
    :param Line line: Spectral line.
    :param Plasma plasma: Plasma object. Default is None.
    :param AtomicData atomic_data: Atomic data provider. Will use `OpenADAS` if None (default).
    :param object lineshape: Spectral line shape model. Will use `GaussianLine` if None (default).
    :param object lineshape_args: Positional arguments for the lineshape.
    :param object lineshape_kwargs: Keyword arguments for the lineshape.
    """

    def __init__(self, SOLPSFunction2D radiation_function, Line line, Plasma plasma=None, AtomicData atomic_data=None,
                 object lineshape=None, object lineshape_args=None, object lineshape_kwargs=None):

        super().__init__(plasma, atomic_data)

        self._line = line

        self._power_density = AxisymmetricMapper(radiation_function)

        self._lineshape_class = lineshape or GaussianLine
        if not issubclass(self._lineshape_class, LineShapeModel):
            raise TypeError("The attribute lineshape must be a subclass of LineShapeModel.")

        if lineshape_args:
            self._lineshape_args = lineshape_args
        else:
            self._lineshape_args = []
        if lineshape_kwargs:
            self._lineshape_kwargs = lineshape_kwargs
        else:
            self._lineshape_kwargs = {}

        # ensure that cache is initialised
        self._change()

    def __repr__(self):
        return '<SOLPSLineEmitter: element={}, charge={}, transition={}>'.format(self._line.element.name, self._line.charge, self._line.transition)

    cpdef Spectrum emission(self, Point3D point, Vector3D direction, Spectrum spectrum):

        cdef double ne, ni, te, radiance

        # cache data on first run
        if self._lineshape is None:
            self._populate_cache()

        # add emission line to spectrum
        radiance = RECIP_4_PI * self._power_density(point.x, point.y, point.z)
        return self._lineshape.add_line(radiance, point, direction, spectrum)

    cdef int _populate_cache(self) except -1:

        # sanity checks
        if self._plasma is None:
            raise RuntimeError("The emission model is not connected to a plasma object.")

        if self._atomic_data is None:
            raise RuntimeError("The emission model is not connected to an atomic data source.")

        if self._line is None:
            raise RuntimeError("The emission line has not been set.")

        # locate target species
        try:
            self._target_species = self._plasma.composition.get(self._line.element, self._line.charge)
        except ValueError:
            raise RuntimeError("The plasma object does not contain the ion species for the specified line "
                               "(element={}, ionisation={}).".format(self._line.element.symbol, self._line.charge))

        # identify wavelength
        self._wavelength = self._atomic_data.wavelength(self._line.element, self._line.charge, self._line.transition)

        # instance line shape renderer
        self._lineshape = self._lineshape_class(self._line, self._wavelength, self._target_species, self._plasma,
                                                *self._lineshape_args, **self._lineshape_kwargs)

    def _change(self):

        # clear cache to force regeneration on first use
        self._target_species = None
        self._wavelength = 0.0
        self._lineshape = None



import unittest

import pickle
import numpy as np
from numpy.random import default_rng

from cherab.solps import SOLPSSimulation
from cherab.solps.tests.test_mesh import create_test_mesh


def create_test_simulation():
    # Creating a test SOLPSSimulation
    mesh, rho, beta, r0 = create_test_mesh()
    nx = mesh.nx
    ny = mesh.ny

    species_list = [('hydrogen', 0), ('hydrogen', 1)]
    sim = SOLPSSimulation(mesh, species_list)
    rng = default_rng(seed=42)
    b0 = 1.
    b_field = np.zeros((3, ny, nx))
    b_field[2] = b0 * r0 / (r0 + 0.5 * (rho[:-1] + rho[1:])[:, None] * np.cos(0.5 * (beta[:-1] + beta[1:]))[None, :])
    b_field[0] = 0.01 * (rng.random((ny, nx)) - 0.5)
    b_field[1] = 0.01 * (rng.random((ny, nx)) - 0.5)
    sim.b_field = b_field
    sim.electron_temperature = 1.e3 * (1 - (rho[:-1, None] - rho[0]) / (rho[-1] - rho[0])) + 1.e2 * (rng.random((ny, nx)) - 0.5)
    sim.electron_density = 1.e19 * (1 - (rho[:-1, None] - rho[0]) / (rho[-1] - rho[0])) + 1.e18 * (rng.random((ny, nx)) - 0.5)
    electron_velocities = 1.e4 * rng.random((3, ny, nx))
    electron_velocities[2] += 1.e5
    sim.electron_velocities = electron_velocities
    sim.ion_temperature = sim.electron_temperature
    sim.neutral_temperature = 0.1 * sim.ion_temperature[None, :, :]
    sim.species_density = np.zeros((2, ny, nx))
    sim.species_density[0] = 0.1 * sim.electron_density[::-1, :]
    sim.species_density[1] = sim.electron_density
    velocities = 1.e4 * rng.random((2, 3, ny, nx))
    velocities[1, 2] -= 1.e5
    sim.velocities = velocities

    return sim


class TestSOLPSSimulation(unittest.TestCase):
    """
    Test SOLPSSimulation.
    """

    sim = create_test_simulation()

    def test_vector_transform(self):
        """ Tests vector transforms (poloidal <--> cylindrical)"""

        test_list = []
        # save a copy of sim.b_field
        b_field = self.sim.b_field
        # this creates a new array at sim.b_field
        self.sim.b_field_cylindrical = self.sim.b_field_cylindrical
        # check if the new array is identical to the original one (poloidal --> cylindrical --> poloidal)
        test_list.append(np.allclose(b_field, self.sim.b_field))

        electron_velocities = self.sim.electron_velocities
        self.sim.electron_velocities_cylindrical = self.sim.electron_velocities_cylindrical
        test_list.append(np.allclose(electron_velocities, self.sim.electron_velocities))

        velocities = self.sim.velocities
        self.sim.velocities_cylindrical = self.sim.velocities_cylindrical
        test_list.append(np.allclose(velocities, self.sim.velocities))

        return self.assertTrue(np.all(test_list))

    def test_pickle(self):
        sim1 = self.sim
        sim2 = pickle.loads(pickle.dumps(sim1))

        test_list = []
        test_list.append(sim1.species_list == sim2.species_list)
        test_list.append(sim1.neutral_list == sim2.neutral_list)

        test_list.append(np.array_equiv(sim1.b_field_cylindrical, sim2.b_field_cylindrical))
        test_list.append(np.array_equiv(sim1.electron_temperature, sim2.electron_temperature))
        test_list.append(np.array_equiv(sim1.ion_temperature, sim2.ion_temperature))
        test_list.append(np.array_equiv(sim1.neutral_temperature, sim2.neutral_temperature))
        test_list.append(np.array_equiv(sim1.electron_density, sim2.electron_density))
        test_list.append(np.array_equiv(sim1.electron_velocities_cylindrical, sim2.electron_velocities_cylindrical))
        test_list.append(np.array_equiv(sim1.species_density, sim2.species_density))
        test_list.append(np.array_equiv(sim1.velocities_cylindrical, sim2.velocities_cylindrical))

        test_list.append(np.allclose(sim1.b_field, sim2.b_field))
        test_list.append(np.allclose(sim1.electron_velocities, sim2.electron_velocities))
        test_list.append(np.allclose(sim1.velocities, sim2.velocities))

        return self.assertTrue(np.all(test_list))


if __name__ == '__main__':
    unittest.main()



import unittest

import pickle
import numpy as np

from cherab.solps import SOLPSMesh


def create_test_mesh():
    # Creating a test SOLPSMesh
    # closed magnetic surfaces only with circular cross section
    r0 = 2.  # major radius
    beta = np.linspace(0, 2 * np.pi, 19)  # poloidal angle
    rho = np.linspace(0.5, 1., 6)  # magnetic surface radius
    # cell vertices
    r2d = r0 + rho[:, None] * np.cos(beta)[None, :]
    z2d = rho[:, None] * np.sin(beta)[None, :]
    # exact cell volume
    vol = (np.pi * r0 * (rho[1:]**2 - rho[:-1]**2)[:, None] * (beta[1:] - beta[:-1])[None, :] +
           2. / 3. * np.pi * (rho[1:]**3 - rho[:-1]**3)[:, None] * (np.sin(beta[1:]) - np.sin(beta[:-1]))[None, :])
    ny = rho.size - 1
    nx = beta.size - 1
    r = np.zeros((4, ny, nx))
    z = np.zeros((4, ny, nx))
    r[0] = r2d[:-1, :-1]
    z[0] = z2d[:-1, :-1]
    r[1] = r2d[:-1, 1:]  # poloidal + 1
    z[1] = z2d[:-1, 1:]  # poloidal + 1
    r[2] = r2d[1:, :-1]  # radial + 1
    z[2] = z2d[1:, :-1]  # radial + 1
    r[3] = r2d[1:, 1:]  # poloidal + 1, radial + 1
    z[3] = z2d[1:, 1:]  # poloidal + 1, radial + 1
    neighbix = -1 * np.ones((4, ny, nx), dtype=int)
    neighbix[0, 1:, :] = np.arange(nx, dtype=int)[None, :]  # left
    neighbix[1, :, 1:] = np.arange(nx - 1, dtype=int)[None, :]  # bottom
    neighbix[2, :-1, :] = np.arange(nx, dtype=int)[None, :]  # right
    neighbix[3, :, :-1] = np.arange(1, nx, dtype=int)[None, :]  # top
    neighbix[1, :, 0] = nx - 1  # closing the surfaces
    neighbix[3, :, -1] = 0  # closing the surfaces
    neighbiy = -1 * np.ones((4, ny, nx), dtype=int)
    neighbiy[0, 1:, :] = np.arange(ny - 1, dtype=int)[:, None]  # left
    neighbiy[1, :, :] = np.arange(ny, dtype=int)[:, None]  # bottom
    neighbiy[2, :-1, :] = np.arange(1, ny, dtype=int)[:, None]  # right
    neighbiy[3, :, :] = np.arange(ny, dtype=int)[:, None]  # top

    mesh = SOLPSMesh(r, z, vol, neighbix, neighbiy)

    return mesh, rho, beta, r0


class TestSOLPSMesh(unittest.TestCase):
    """
    Test SOLPSMesh.
    """

    mesh, _, _, _ = create_test_mesh()

    def test_pickle(self):
        """ Tests SOLPSMesh pickling"""

        mesh1 = self.mesh
        mesh2 = pickle.loads(pickle.dumps(mesh1))
        test_list = []
        # test initial parameters
        test_list.append(np.array_equiv(mesh1.r, mesh2.r))
        test_list.append(np.array_equiv(mesh1.z, mesh2.z))
        test_list.append(np.array_equiv(mesh1.vol, mesh2.vol))
        test_list.append(np.array_equiv(mesh1.neighbix, mesh2.neighbix))
        test_list.append(np.array_equiv(mesh1.neighbiy, mesh2.neighbiy))
        # test derivative parameters
        test_list.append(np.array_equiv(mesh1.radial_basis_vector, mesh2.radial_basis_vector))
        test_list.append(np.array_equiv(mesh1.poloidal_basis_vector, mesh2.poloidal_basis_vector))
        test_list.append(np.array_equiv(mesh1.vertex_coordinates, mesh2.vertex_coordinates))
        test_list.append(np.array_equiv(mesh1.quadrangles, mesh2.quadrangles))

        return self.assertTrue(np.all(test_list))


if __name__ == '__main__':
    unittest.main()



# Copyright 2016-2018 Euratom
# Copyright 2016-2018 United Kingdom Atomic Energy Authority
# Copyright 2016-2018 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.

import numpy as np
from cherab.solps.eirene import Eirene
from cherab.solps.eirene.parser.utility import read_labelled_block44


def load_fort44_post_2016(file_path, debug=False):
    """
    Read neutral species and wall flux information from fort.44.

    This is for fort.44 files with format ID 20170328.
    Specification of the data format is in Section 5.2 of the SOLPS
    manual (Running the coupled version -> Output Files).

    :param str file_path: path to the fort.44 file
    :param bool debug: status flag for printing debugging output
    :rtype:
    """

    with open(file_path, 'r') as file_handle:

        # Read sizes
        line = file_handle.readline().split()
        nx = int(line[0])
        ny = int(line[1])
        version = int(line[2])
        if debug:
            print('Geometry & Version : nx {}, ny {}, version {}'
                  .format(nx, ny, version))

        # Read Species numbers
        line = file_handle.readline().split()
        na = int(line[0])  # number of atoms
        nm = int(line[1])  # number of molecules
        ni = int(line[2])  # number of ions
        ns = na + nm + ni  # total number of species

        if debug:
            print('Species # : {} atoms, {} molecules, {} ions, {} total species'
                  .format(na, nm, ni, ns))

        # Read Species labels
        species_labels = []
        for _ in range(ns):
            line = file_handle.readline()
            species_labels.append(line.strip())
        if debug:
            print("Species labels => {}".format(species_labels))

        # Read all blocks of data from the file into a dictionary.
        raw_data = {}
        while True:
            try:
                raw_data.update(read_labelled_block44(file_handle))
            except EOFError:
                break

    # create eirene object
    eirene = Eirene(nx, ny, na, nm, ni, ns, species_labels, version)

    # Read atomic species data
    eirene.da = raw_data['dab2'].reshape(eirene.na, eirene.ny, eirene.nx)
    eirene.ta = raw_data['tab2'].reshape(eirene.na, eirene.ny, eirene.nx)
    eirene.rpa = raw_data['rfluxa'].reshape(eirene.na, eirene.ny, eirene.nx)
    eirene.ppa = raw_data['pfluxa'].reshape(eirene.na, eirene.ny, eirene.nx)
    eirene.rea = raw_data['refluxa'].reshape(eirene.na, eirene.ny, eirene.nx)
    eirene.pea = raw_data['pefluxa'].reshape(eirene.na, eirene.ny, eirene.nx)

    # Read molecular species data
    if eirene.nm > 0:
        eirene.dm = raw_data['dmb2'].reshape(eirene.nm, eirene.ny, eirene.nx)
        eirene.tm = raw_data['tmb2'].reshape(eirene.nm, eirene.ny, eirene.nx)
        eirene.rpm = raw_data['rfluxm'].reshape(eirene.nm, eirene.ny, eirene.nx)
        eirene.ppm = raw_data['pfluxm'].reshape(eirene.nm, eirene.ny, eirene.nx)
        eirene.rem = raw_data['refluxm'].reshape(eirene.nm, eirene.ny, eirene.nx)
        eirene.pem = raw_data['pefluxm'].reshape(eirene.nm, eirene.ny, eirene.nx)
        eirene.emism = raw_data['emissmol'].reshape(1, eirene.ny, eirene.nx)
        eirene.edism = raw_data['edissml'].reshape(eirene.nm, eirene.ny, eirene.nx)

    # Read ion species data
    if eirene.ni > 0:
        eirene.di = raw_data['dib2'].reshape(eirene.ni, eirene.ny, eirene.nx)
        eirene.ti = raw_data['tib2'].reshape(eirene.ni, eirene.ny, eirene.nx)

    # Halpha emission
    # fort.44 gives emission due to atoms and emission due to molecules separately.
    emiss_atoms = raw_data['emiss'].reshape(1, eirene.ny, eirene.nx)
    if eirene.nm > 0:
        eirene.emist = emiss_atoms + eirene.emism
    else:
        eirene.emist = emiss_atoms

    # Radiated power
    try:
        eneutrad = raw_data['eneutrad'].reshape(eirene.na, eirene.ny, eirene.nx)
    except ValueError:
        # Some Eirene/SOLPS combinations output eneutrad summed over atoms,
        # rather than resolved.
        eneutrad = raw_data['eneutrad'].reshape(1, eirene.ny, eirene.nx)

    try:
        emolrad = raw_data['emolrad'].reshape(eirene.nm, eirene.ny, eirene.nx)
    except KeyError:
        print("Warning: no emolrad data available in EIRENE simulation.")
        emolrad = 0
    except ValueError:
        # Data is summed over molecules rather than resolved.
        emolrad = raw_data['emolrad'].reshape(1, eirene.ny, eirene.nx)

    try:
        eionrad = raw_data['eionrad'].reshape(eirene.ni, eirene.ny, eirene.nx)
    except KeyError:
        print("Warning: no eionrad data available in EIRENE simulation.")
        eionrad = 0
    except ValueError:
        # Data is summed over ions rather than resolved.
        eionrad = raw_data['eionrad'].reshape(1, eirene.ny, eirene.nx)

    eirene.eradt = (np.sum(eneutrad, axis=0, keepdims=True)
                    + np.sum(emolrad, axis=0, keepdims=True)
                    + np.sum(eionrad, axis=0, keepdims=True))

    return eirene




# Copyright 2016-2018 Euratom
# Copyright 2016-2018 United Kingdom Atomic Energy Authority
# Copyright 2016-2018 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.


import os
import numpy as np


SIM_INFO_DATA = 0
MESH_DATA = 1


# Code based on a script by Felix Reimold (2016)

def load_b2f_file(filepath, debug=False, header_dict=None):
    """
    File for parsing the 'b2fstate and b2fplasmf' B2 Eirene output files.

    :param str filepath: full path for file to load and parse
    :param bool debug: flag for displaying textual debugging information.
    :param dict header_dict: header information
    :return: tuple of dictionaries. First is the header information such
      as the version, label, grid size, etc. Second contains arrays of
      simulation outputs. Third contains information about the simulation
      itself.

    If header_dict is provided, it should contain the dimension info 'nx',
    'ny', 'nxg', 'nyg', 'nxy', 'nxyg', as obtained by reading a previous
    b2f* file. If header_dict is None, these values will be read from the
    top of the file.
    """

    # Inline function for mapping str data to floats, reshaping arrays, and loading into SOLPSData object.
    def _make_solps_data_object(_data):

        # Convert list of strings to list of floats
        for idx, item in enumerate(_data):
            _data[idx] = float(item)

        # Multiple 2D data field (e.g. na)
        if number > nxyg:
            _data = np.array(_data).reshape((number // nxyg, nyg, nxg))
            if debug:
                print('Mesh data field {} with dimensions:  {:d} x {:d} x {:d}'.format(name, number // nxyg, nyg, nxg))
            return MESH_DATA, _data

        # 2D data field (e.g. ne)
        elif number == nxyg:
            _data = np.array(_data).reshape((nyg, nxg))
            if debug:
                print('Mesh data field {} with dimensions:  {:d} x {:d}'.format(name, nyg, nxg))
            return MESH_DATA, _data

        # Additional information field (e.g. zamin)
        else:
            _data = np.array(_data)
            if debug:
                print('Sim info field {} with length:     {} '.format(name, _data.shape[0]))
            return SIM_INFO_DATA, _data

    if not(os.path.isfile(filepath)):
        raise IOError('File {} not found.'.format(filepath))

    # Open SOLPS geometry file for reading
    fh = open(filepath, 'r')

    # Version header
    version = fh.readline()

    if header_dict is None:
        # Read mesh size
        fh.readline()
        line = fh.readline().split()
        nx = int(line[0])
        ny = int(line[1])

        # Calculate guard cells
        nxg = nx + 2
        nyg = ny + 2

        # Flat vector size
        nxy = nx * ny
        nxyg = nxg * nyg

        # Read Label
        fh.readline()
        label = fh.readline()

        # Save header
        header_dict = {'version': version, 'nx': nx, 'ny': ny, 'nxg': nxg, 'nyg': nyg, 'nxy': nxy, 'nxyg': nxyg, 'label': label}
    else:
        nx = header_dict['nx']
        ny = header_dict['ny']
        nxg = header_dict['nxg']
        nyg = header_dict['nyg']
        nxy = header_dict['nxy']
        nxyg = header_dict['nxyg']

    # variables for file data
    name = ''
    number = 0
    data = []
    sim_info_dict = {}
    mesh_data_dict = {}

    # Repeat reading data blocks till EOF
    while True:
        # Read line
        line = fh.readline().split()

        # EOF --OR--  New block of similar data (vector qty, e.g. bb)
        if len(line) == 0:
            # Check if last line
            line = fh.readline().split()
            if len(line) == 0:
                break

        # New block found
        if line[0] == '*cf:':

            # If previous block read --> Convert data to float, reshape and save to Object
            if name != '':
                flag, shaped_data = _make_solps_data_object(data)
                if flag == SIM_INFO_DATA:
                    sim_info_dict[name] = shaped_data
                elif flag == MESH_DATA:
                    mesh_data_dict[name] = shaped_data

            # Read next field paramters
            data_type = str(line[1].strip())
            number = int(line[2])
            name = str(line[3].strip())
            data = []

        # Append line to vector of data
        else:
            data.extend(line)

    if name != '':
        flag, shaped_data = _make_solps_data_object(data)
        if flag == SIM_INFO_DATA:
            sim_info_dict[name] = shaped_data
        elif flag == MESH_DATA:
            mesh_data_dict[name] = shaped_data

    return header_dict, sim_info_dict, mesh_data_dict


# Test case
if __name__ == '__main__':

    p = load_b2f_file("/home/mcarr/mst1/aug_2016/solps_testcase/b2fstate", debug=True)



# Copyright 2016-2018 Euratom
# Copyright 2016-2018 United Kingdom Atomic Energy Authority
# Copyright 2016-2018 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.

from cherab.solps.eirene import Eirene
from cherab.solps.eirene.parser.utility import read_unlabelled_block44 as read_block44


def load_fort44_2013(file_path, debug=False):
    """ Read neutral species and wall flux information from fort.44

    Template for reading is ngread.F in b2plot of B2.5 source.
    This is for fort.44 files with format ID 20130210.

    :param str file_path: path to the fort.44 file
    :param bool debug: status flag for printing debugging output
    :rtype:
    """

    with open(file_path, 'r') as file_handle:
        # Read sizes
        line = file_handle.readline().split()
        nx = int(line[0])
        ny = int(line[1])
        version = int(line[2])
        if debug:
            print('Geometry & Version : nx {}, ny {}, version {}'
                  .format(nx, ny, version))

        # Read Species numbers
        line = file_handle.readline().split()
        na = int(line[0])  # number of atoms
        nm = int(line[1])  # number of molecules
        ni = int(line[2])  # number of ions
        ns = na + nm + ni  # total number of species
        if debug:
            print('Species # : {} atoms, {} molecules, {} ions, {} total species'
                  .format(na, nm, ni, ns))

        # Read Species labels
        species_labels = []
        for _ in range(ns):
            line = file_handle.readline()
            species_labels.append(line.strip())
        if debug:
            print("Species labels => {}".format(species_labels))

        # create eirene object
        eirene = Eirene(nx, ny, na, nm, ni, ns, species_labels, version)

        # Read atomic species (da, ta)
        eirene.da = read_block44(file_handle, eirene.na, nx, ny)  # Atomic Neutral Density
        eirene.ta = read_block44(file_handle, eirene.na, nx, ny)  # Atomic Neutral Temperature
        if debug:
            print('Atomic Neutral Density nD0: ', eirene.da[0, :, 0])
            print('Atomic Neutral Temperature TD0: ', eirene.ta[0, :, 0])

        # Read molecular species (dm, tm)
        eirene.dm = read_block44(file_handle, eirene.nm, nx, ny)  # Molecular Neutral Density
        eirene.tm = read_block44(file_handle, eirene.nm, nx, ny)  # Molecular Neutral Temperature

        # Read ion species (di, ti)
        eirene.di = read_block44(file_handle, eirene.ni, nx, ny)  # Test Ion Density
        eirene.ti = read_block44(file_handle, eirene.ni, nx, ny)  # Test Ion Temperature

        # Read radial particle flux (rpa, rpm)
        eirene.rpa = read_block44(file_handle, eirene.na, nx, ny)  # Atomic Radial Particle Flux
        eirene.rpm = read_block44(file_handle, eirene.nm, nx, ny)  # Molecular Radial Particle Flux

        # Read poloidal particle flux (ppa, ppm)
        eirene.ppa = read_block44(file_handle, eirene.na, nx, ny)  # Atomic Poloidal Particle Flux
        eirene.ppm = read_block44(file_handle, eirene.nm, nx, ny)  # Molecular Poloidal Particle Flux

        # Read radial energy flux (rea, rem)
        eirene.rea = read_block44(file_handle, eirene.na, nx, ny)  # Atomic Radial Energy Flux
        eirene.rem = read_block44(file_handle, eirene.nm, nx, ny)  # Molecular Radial Energy Flux

        # Read poloidal energy flux (pea, pem)
        eirene.pea = read_block44(file_handle, eirene.na, nx, ny)  # Atomic Poloidal Energy Flux
        eirene.pem = read_block44(file_handle, eirene.nm, nx, ny)  # Molecular Poloidal Energy Flux

        # Halpha total & molecules (emist, emism)
        eirene.emist = read_block44(file_handle, 1, nx, ny)  # Total Halpha Emission (including molecules)
        eirene.emism = read_block44(file_handle, 1, nx, ny)  # Molecular Halpha Emission

        # Radiated power (elosm, edism, eradt)
        eirene.elosm = read_block44(file_handle, nm, nx, ny)  # Power loss due to molecules (including dissociation)
        eirene.edism = read_block44(file_handle, nm, nx, ny)  # Power loss due to molecule dissociation
        eirene.eradt = read_block44(file_handle, 1, nx, ny)  # Neutral radiated power

    return eirene



from setuptools import setup, find_packages, Extension
from Cython.Build import cythonize
import sys
import numpy
import os
import os.path as path

force = False
profile = False

if "--force" in sys.argv:
    force = True
    del sys.argv[sys.argv.index("--force")]

if "--profile" in sys.argv:
    profile = True
    del sys.argv[sys.argv.index("--profile")]

compilation_includes = [".", numpy.get_include()]

setup_path = path.dirname(path.abspath(__file__))

# build extension list
extensions = []
for root, dirs, files in os.walk(setup_path):
    for file in files:
        if path.splitext(file)[1] == ".pyx":
            pyx_file = path.relpath(path.join(root, file), setup_path)
            module = path.splitext(pyx_file)[0].replace("/", ".")
            extensions.append(Extension(module, [pyx_file], include_dirs=compilation_includes),)

cython_directives = {"language_level": 3}
if profile:
    cython_directives["profile"] = True


with open("README.md") as f:
    long_description = f.read()

setup(
    name="cherab-solps",
    version="1.2.0rc2",
    license="EUPL 1.1",
    namespace_packages=['cherab'],
    description="Cherab spectroscopy framework: SOLPS submodule",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Education",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Cython",
        "Programming Language :: Python :: 3",
        "Topic :: Scientific/Engineering :: Physics",
    ],
    url="https://github.com/cherab",
    project_urls=dict(
        Tracker="https://github.com/cherab/solps/issues",
        Documentation="https://cherab.github.io/documentation/",
    ),
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=find_packages(),
    include_package_data=True,
    install_requires=["raysect==0.7.1", "cherab==1.3"],
    ext_modules=cythonize(extensions, force=force, compiler_directives=cython_directives),
)
